/**
 * Programm that implements the basic FIFO-operations and tests them (init, empty, full, enqueue, dequeue)
 * 
 * @author robert.binna@uibk.ac.at
 * @date 09.04.2008
 */
 
#include <stdio.h>
#include <assert.h>
 
#define BUFFERSIZE 30

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif
 
int first = 0; /* first index, the index where to insert the next element enqueue */
int last = 0; /* last index, the index where the next element is returned dequeue */
int isFull = 0;

int fifo_buffer[BUFFERSIZE];

 
/**
 * Initializes the FIFO / resets it
 */
void init() {
	first = 0;
	last = 0;
	isFull = 0;
}

/**
 * Checks if the FIFO is empty at the call time
 *
 * @return if it is empty
 */
int empty() {
	return !isFull && first == last;  
}

/**
 * Checks if the FIFO is full and nothing can get inserted anymore
 *
 * @return if it is full
 */
int full() {
	return isFull;
}

/**
 * Dequeue adds an element to the tail of the queue. It is responsible for returning an error if
 * the queue is full
 *
 * @returns 1 if all is okay 0 if the the FIFO is full
 */
int enqueue(int element) {
	int success = TRUE;
	if(full()) {
		success = FALSE;
	} else {
		fifo_buffer[first] = element;
		++first;
		first%=BUFFERSIZE;
		isFull = (first == last);
	}
	return success;
}

/**
 * Dequeue removes the first element from the fifo queue. It is responsible for returning an error if
 * the queue is empty
 *
 * @param output an output parameter. It will be filled with the element that is removed from 
 * the FIFO. If the buffer is empty the value will be the same as before
 * @return 1 if removal of the first element was okay otherwise 0. 0 indicates that the FIFO was already empty.
 */
int dequeue(int* output) {
	int success = TRUE;
	if(empty()) {
		success = FALSE;
	} else {
		*output = fifo_buffer[last];
		++last;
		last%=BUFFERSIZE;
		isFull = FALSE;
	}
	return success;
}

/**
 * main function for testing the fifo operations
 * tests the correct behaviour of all base operations. The behaviour is checked through 
 * assert macro calls.
 * @return the exit value of the program. 
 * 0 for correct behaviour.
 */
int main(void) {
	int i=0;

	init();

	printf("FIFO empty (should be 1): %d\n", empty());
	assert(empty());
	printf("FIFO full  (should be 0): %d\n", full());
	assert(!full());
	
	for(i=0; i < BUFFERSIZE; ++i) {
		assert(!full());
		printf("Enqueue %d\n",i);
		assert(enqueue(i));
		assert(!empty());
	}
	
	assert(full());
	assert(!enqueue(i));
	
	for(i=0; i < BUFFERSIZE; ++i) {
		assert(!empty());
		int output = 0;
		printf("Dequeue %d",i);
		assert(dequeue(&output));
		printf("Got Value %d\n",output);
		assert(output == i);
		assert(!full());
	}
	
	printf("FIFO empty (should be 1): %d\n", empty());
	assert(empty());
	printf("FIFO full  (should be 0): %d\n", full());
	assert(!full());
	
	assert(!dequeue(&i));
	
	init();

	printf("FIFO empty (should be 1): %d\n", empty());
	assert(empty());
	printf("FIFO full  (should be 0): %d\n", full());
	assert(!full());
	
	for(i=0; i < BUFFERSIZE; ++i) {
		assert(!full());
		printf("Enqueue %d\n",i);
		assert(enqueue(i));
		assert(!empty());
	}
	
	assert(full());
	assert(!enqueue(i));
	
	init();
	
	printf("FIFO empty (should be 1): %d\n", empty());
	assert(empty());
	printf("FIFO full  (should be 0): %d\n", full());
	assert(!full());
	
	return 0;
}
